#include "analyze_argument.h"
#include "check_target.h"
#include "do_compare_verify_two.h"	// inline
#include "get_drive_list.h"
#include "get_settings_dir.h"
#include "help.h"
#include "print_error.h"	// inline
#include "read_conf.h"
#include "set_signal.h"
#include "struct_CPDD.h"
#include "struct_DPath.h"
#include "struct_OPArg.h"
#include "verify_hash.h"
#include "write_conf.h"

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>

/* 関数プロトタイプ */
static inline void delete_src(const OPArg *oparg, const VList *file_list, CPDD *cpdd) __attribute__((always_inline));

/*******************************************************************************
*******************************************************************************/
int main(int argc, char **argv)
{
	if(argc == 1)
	{
		help();
		return EXIT_SUCCESS;
	}

	SDir *sdir = get_settings_dir();

	if(argc == 2)
	{
		char *tmp = argv[1];

		/* -oの場合、オプションを設定、保存し、終了する。 */
		if((tmp[0] == '-') && (tmp[1] == 'o'))
		{
			write_conf(sdir);
			return EXIT_SUCCESS;
		}
		/* -hの場合はヘルプを表示する */
		else if((tmp[0] == '-') && (tmp[1] == 'h'))
		{
			help();
			return EXIT_SUCCESS;
		}
		/* --helpの場合もヘルプを表示する */
		else if((tmp[0] == '-') && (tmp[1] == '-') && (tmp[2] == 'h') && (tmp[3] == 'e') && (tmp[4] == 'l') && (tmp[5] == 'p'))
		{
			help();
			return EXIT_SUCCESS;
		}
		else
		{
			fprintf(stderr, "引数が足りません\n");
			return EXIT_FAILURE;
		}
	}

	OPArg *oparg = read_conf(sdir);
	analyze_argument(argc, argv, oparg);
	VList *drive_list = get_drive_list();
	set_signal();

	#ifdef _DEBUG_FLAG
		puts("Debug modeで動作しています...");
		fflush(stdout);
	#endif

	puts("コピーを開始します...");
	fflush(stdout);

	RData *rdata = check_target(oparg, drive_list, sdir);

	puts("コピーを完了しました...");

	VList *file_list = rdata->file_list;
	CPDD *cpdd = rdata->cpdd;

	if((file_list != NULL) && (oparg->CHECK_MODE == AFTER) && (cpdd->mk_file_count > 0))
	{
		puts("ハッシュ値のチェックを開始します...");
		fflush(stdout);

		do_compare_verify_two(oparg, file_list, sdir, cpdd);

		puts("ハッシュ値のチェックを完了しました...");
		fflush(stdout);
	}

	if(oparg->M == MOVE)
	{
		if(
			((oparg->CHECK == NOT) && ((cpdd->mk_file_count > 0) || (cpdd->mk_link_count > 0) || (cpdd->dir_count > 0)))
			||
			((oparg->CHECK != NOT) && ((cpdd->mk_link_count > 0) || (cpdd->dir_count > 0)))
		)
		{
			puts("コピー元の削除を開始します...");
			fflush(stdout);

			delete_src(oparg, file_list, cpdd);

			puts("コピー元の削除を完了しました...");
			fflush(stdout);
		}
	}

	puts("");

	if(cpdd->file_count > 0)
	{
		printf("File   : %lld / %lld", cpdd->mk_file_count, cpdd->file_count);

		if(cpdd->thread_mode_count > 0) { printf(" (Thread : %lld / %lld)\n", cpdd->thread_mode_count, cpdd->mk_file_count); }
		else { puts(""); }
	}

	if(cpdd->link_count > 0)   { printf("Link   : %lld / %lld\n", cpdd->mk_link_count, cpdd->link_count); }
	if(cpdd->dir_count > 0)    { printf("Dir    : %lld / %lld\n", cpdd->mk_dir_count, cpdd->dir_count); }
	if(cpdd->total_source > 0) { printf("Source : %lld MB (%lld Bytes)\n", cpdd->total_source / (1024 * 1024), cpdd->total_source); }
	if(cpdd->total_read > 0)   { printf("Read   : %lld MB (%lld Bytes)\n", cpdd->total_read / (1024 * 1024), cpdd->total_read); }
	if(cpdd->total_write > 0)  { printf("Write  : %lld MB (%lld Bytes)\n", cpdd->total_write / (1024 * 1024), cpdd->total_write); }
	if(cpdd->total_error > 0)  { printf("Error  : %lld\n", cpdd->total_error); }

	/*
	 * close()する必要があるもの
	 * sdir->current_fd
	 */

	/* 基本的に、mallocしたのはfreeしてない。 */

	puts("");

	return EXIT_SUCCESS;
}

/*******************************************************************************
*******************************************************************************/
#include <unistd.h>

/*******************************************************************************
*******************************************************************************/
static inline void delete_src(const OPArg *oparg, const VList *file_list, CPDD *cpdd)
{
	/*
	 * キャストしないと警告が出る
	 * warning: initialization discards qualifiers from pointer target type
	 */
	VList *tmp = (VList *)file_list;
	tmp = tmp->tail;

	for(;;)
	{
		CPInfo *cpinfo = (CPInfo *)tmp->data;

		switch(cpinfo->src_type)
		{
		case REGULAR:
			if(oparg->CHECK == NOT)
			{
				if((cpinfo->d_para == WRITE) && (cpinfo->src_size <= cpinfo->dst_size))
				{
					SOURCE_UNLINK
				}
			}
			else
			{
				if((cpinfo->hash_check == true) && (cpinfo->d_para == WRITE) && (cpinfo->s_para != UNLINK) && (cpinfo->src_size <= cpinfo->dst_size))
				{
					SOURCE_UNLINK
				}
				else if((cpinfo->d_para == WRITE) && (cpinfo->s_para != UNLINK) && (cpinfo->src_size == 0) && (cpinfo->dst_size == 0))
				{
					SOURCE_UNLINK
				}
			}
			break;

		case SYMBOLICLINK:
			if(cpinfo->d_para == WRITE)
			{
				SOURCE_UNLINK
			}
			break;

		case DIRECTORY:
			errno = 0;
			if(rmdir(cpinfo->src) == -1)
			{
				if((errno != ENOTEMPTY) && (errno != ENOENT))
				{
					print_error("remove", __FILE__, __LINE__, cpdd);
					fprintf(stderr, "`%s' の削除に失敗しました\n", cpinfo->src);
				}
			}
			break;

		default:
			fprintf(stderr, "ファイルタイプが不正です\n");
			break;
		}

		if(tmp->prev == NULL)
		{
			break;
		}

		tmp = tmp->prev;
	}
}
